@@ -1,13 +1,12 @@
 # alias to keep the 'bytecode' variable free
 import bytecode as _bytecode
-from bytecode.instr import Instr, Label, UNSET
+from bytecode.instr import UNSET, Label, SetLineno, Instr
 
 
 class BaseBytecode:
     def __init__(self):
         self.argcount = 0
-        # FIXME: rename to kwonlyargcount?
-        self.kw_only_argcount = 0
+        self.kwonlyargcount = 0
         # FIXME: insane and safe value until _ConvertBytecodeToConcrete is able
         # to compute the value itself
         self._stacksize = 256
@@ -24,7 +23,7 @@ class BaseBytecode:
 
     def _copy_attr_from(self, bytecode):
         self.argcount = bytecode.argcount
-        self.kw_only_argcount = bytecode.kw_only_argcount
+        self.kwonlyargcount = bytecode.kwonlyargcount
         self._stacksize = bytecode._stacksize
         self.flags = bytecode.flags
         self.first_lineno = bytecode.first_lineno
@@ -40,7 +39,7 @@ class BaseBytecode:
 
         if self.argcount != other.argcount:
             return False
-        if self.kw_only_argcount != other.kw_only_argcount:
+        if self.kwonlyargcount != other.kwonlyargcount:
             return False
         if self._stacksize != other._stacksize:
             return False
@@ -68,21 +67,22 @@ class _InstrList(list):
         labels = {}
         jumps = []
 
+        offset = 0
         for index, instr in enumerate(self):
             if isinstance(instr, Label):
-                name = 'label_instr%s' % index
-                instructions.append(name)
-                labels[instr] = name
+                instructions.append('label_instr%s' % index)
+                labels[instr] = offset
             else:
-                if isinstance(instr.arg, Label):
-                    # copy the instruction to be able to modify
-                    # its argument above
-                    instr = instr.copy()
-                    jumps.append(instr)
+                if isinstance(instr, Instr) and isinstance(instr.arg, Label):
+                    target_label = instr.arg
+                    instr = _bytecode.ConcreteInstr(instr.name, 0,
+                                                    lineno=instr.lineno)
+                    jumps.append((target_label, instr))
                 instructions.append(instr)
+                offset += 1
 
-        for instr in jumps:
-            instr.arg = labels[instr.arg]
+        for target_label, instr in jumps:
+            instr.arg = labels[target_label]
 
         return instructions
 
@@ -100,6 +100,18 @@ class Bytecode(_InstrList, BaseBytecode):
             self.extend(instructions)
         self.argnames = []
 
+    def __iter__(self):
+        instructions = super().__iter__()
+        for instr in instructions:
+            if not isinstance(instr, (Label, SetLineno, Instr,
+                                      _bytecode.ConcreteInstr)):
+                raise ValueError("Bytecode must only contain Label, "
+                                 "SetLineno, Instr and ConcreteInstr objects, "
+                                 "but %s was found"
+                                 % instr.__class__.__name__)
+
+            yield instr
+
     @staticmethod
     def from_code(code):
         concrete = _bytecode.ConcreteBytecode.from_code(code)
