@@ -1,10 +1,11 @@
 #!/usr/bin/env python3
 import opcode
+import sys
 import types
 import unittest
 from bytecode import (UNSET, Label, Instr, SetLineno, Bytecode,
+                      CellVar, FreeVar,
                       ConcreteInstr, ConcreteBytecode)
-from bytecode.concrete import ARG_MAX
 from bytecode.tests import get_code, TestCase
 
 
@@ -29,14 +30,14 @@ class ConcreteInstrTests(TestCase):
 
         # test maximum argument
         with self.assertRaises(ValueError):
-            ConcreteInstr("LOAD_CONST", ARG_MAX + 1)
-        instr = ConcreteInstr("LOAD_CONST", ARG_MAX)
-        self.assertEqual(instr.arg, ARG_MAX)
+            ConcreteInstr("LOAD_CONST", 2147483647 + 1)
+        instr = ConcreteInstr("LOAD_CONST", 2147483647)
+        self.assertEqual(instr.arg, 2147483647)
 
     def test_attr(self):
         instr = ConcreteInstr("LOAD_CONST", 5, lineno=12)
         self.assertEqual(instr.name, 'LOAD_CONST')
-        self.assertEqual(instr.op, 100)
+        self.assertEqual(instr.opcode, 100)
         self.assertEqual(instr.arg, 5)
         self.assertEqual(instr.lineno, 12)
         self.assertEqual(instr.size, 3)
@@ -47,20 +48,18 @@ class ConcreteInstrTests(TestCase):
         instr.set('NOP')
         self.assertEqual(instr.name, 'NOP')
         self.assertIs(instr.arg, UNSET)
-        self.assertIsNone(instr.lineno)
+        self.assertEqual(instr.lineno, 3)
 
-        instr.set('LOAD_FAST', 8, lineno=9)
+        instr.set('LOAD_FAST', 8)
         self.assertEqual(instr.name, 'LOAD_FAST')
         self.assertEqual(instr.arg, 8)
-        self.assertEqual(instr.lineno, 9)
+        self.assertEqual(instr.lineno, 3)
 
         # invalid
         with self.assertRaises(ValueError):
             instr.set('LOAD_CONST')
         with self.assertRaises(ValueError):
             instr.set('NOP', 5)
-        with self.assertRaises(ValueError):
-            instr.set('LOAD_CONST', 5, lineno=-1)
 
     def test_set_attr(self):
         instr = ConcreteInstr("LOAD_CONST", 5, lineno=12)
@@ -68,16 +67,16 @@ class ConcreteInstrTests(TestCase):
         # operator name
         instr.name = 'LOAD_FAST'
         self.assertEqual(instr.name, 'LOAD_FAST')
-        self.assertEqual(instr.op, 124)
+        self.assertEqual(instr.opcode, 124)
         self.assertRaises(TypeError, setattr, instr, 'name', 3)
         self.assertRaises(ValueError, setattr, instr, 'name', 'xxx')
 
         # operator code
-        instr.op = 100
+        instr.opcode = 100
         self.assertEqual(instr.name, 'LOAD_CONST')
-        self.assertEqual(instr.op, 100)
-        self.assertRaises(ValueError, setattr, instr, 'op', -12)
-        self.assertRaises(TypeError, setattr, instr, 'op', 'abc')
+        self.assertEqual(instr.opcode, 100)
+        self.assertRaises(ValueError, setattr, instr, 'opcode', -12)
+        self.assertRaises(TypeError, setattr, instr, 'opcode', 'abc')
 
         # extended argument
         instr.arg = 0x1234abcd
@@ -91,7 +90,7 @@ class ConcreteInstrTests(TestCase):
 
         # invalid argument
         self.assertRaises(ValueError, setattr, instr, 'arg', -1)
-        self.assertRaises(ValueError, setattr, instr, 'arg', ARG_MAX + 1)
+        self.assertRaises(ValueError, setattr, instr, 'arg', 2147483647 + 1)
 
         # size attribute is read-only
         self.assertRaises(AttributeError, setattr, instr, 'size', 3)
@@ -152,29 +151,69 @@ class ConcreteBytecodeTests(TestCase):
                               ConcreteInstr('RETURN_VALUE', lineno=1)])
         # FIXME: test other attributes
 
+    def test_iter_invalid_types(self):
+        code = ConcreteBytecode()
+        code.append(Label())
+        with self.assertRaises(ValueError):
+            list(code)
+
     def test_to_code_lnotab(self):
         # x = 7
         # y = 8
         # z = 9
-        code = ConcreteBytecode()
-        code.consts = [7, 8, 9]
-        code.names = ['x', 'y', 'z']
-        code.extend([ConcreteInstr("LOAD_CONST", 0, lineno=1),
-                     ConcreteInstr("STORE_NAME", 0, lineno=1),
-                     ConcreteInstr("LOAD_CONST", 1, lineno=2),
-                     ConcreteInstr("STORE_NAME", 1, lineno=2),
-                     ConcreteInstr("LOAD_CONST", 2, lineno=3),
-                     ConcreteInstr("STORE_NAME", 2, lineno=3)])
+        concrete = ConcreteBytecode()
+        concrete.consts = [7, 8, 9]
+        concrete.names = ['x', 'y', 'z']
+        concrete.first_lineno = 3
+        concrete.extend([ConcreteInstr("LOAD_CONST", 0),
+                         ConcreteInstr("STORE_NAME", 0),
+                         SetLineno(4),
+                         ConcreteInstr("LOAD_CONST", 1),
+                         ConcreteInstr("STORE_NAME", 1),
+                         SetLineno(5),
+                         ConcreteInstr("LOAD_CONST", 2),
+                         ConcreteInstr("STORE_NAME", 2)])
 
-        code_obj = code.to_code()
+        code = concrete.to_code()
         expected = (b'd\x00\x00'
                     b'Z\x00\x00'
                     b'd\x01\x00'
                     b'Z\x01\x00'
                     b'd\x02\x00'
                     b'Z\x02\x00')
-        self.assertEqual(code_obj.co_code, expected)
-        self.assertEqual(code_obj.co_lnotab, b'\x06\x01\x06\x01')
+        self.assertEqual(code.co_code, expected)
+        self.assertEqual(code.co_firstlineno, 3)
+        self.assertEqual(code.co_lnotab, b'\x06\x01\x06\x01')
+
+    def test_negative_lnotab(self):
+        # x = 7
+        # y = 8
+        concrete = ConcreteBytecode()
+        concrete.consts = [7, 8]
+        concrete.names = ['x', 'y']
+        concrete.first_lineno = 5
+        concrete.extend([ConcreteInstr("LOAD_CONST", 0),
+                         ConcreteInstr("STORE_NAME", 0),
+                         # line number goes backward!
+                         SetLineno(2),
+                         ConcreteInstr("LOAD_CONST", 1),
+                         ConcreteInstr("STORE_NAME", 1)])
+
+        if sys.version_info >= (3, 6):
+            code = concrete.to_code()
+            expected = (b'd\x00\x00'
+                        b'Z\x00\x00'
+                        b'd\x01\x00'
+                        b'Z\x01\x00')
+            self.assertEqual(code.co_code, expected)
+            self.assertEqual(code.co_firstlineno, 5)
+            self.assertEqual(code.co_lnotab, b'\x06\xfd')
+        else:
+            with self.assertRaises(ValueError) as cm:
+                code = concrete.to_code()
+            self.assertEqual(str(cm.exception),
+                             "negative line number delta is not supported "
+                             "on Python < 3.6")
 
     def test_to_bytecode_consts(self):
         # x = -0.0
@@ -220,7 +259,7 @@ class ConcreteBytecodeTests(TestCase):
         bytecode = concrete.to_bytecode()
         self.assertEqual(bytecode.cellvars, ['x'])
         self.assertEqual(list(bytecode),
-                         [Instr('LOAD_DEREF', 'x', lineno=1)])
+                         [Instr('LOAD_DEREF', CellVar('x'), lineno=1)])
 
     def test_freevar(self):
         concrete = ConcreteBytecode()
@@ -237,7 +276,7 @@ class ConcreteBytecodeTests(TestCase):
         bytecode = concrete.to_bytecode()
         self.assertEqual(bytecode.cellvars, [])
         self.assertEqual(list(bytecode),
-                         [Instr('LOAD_DEREF', 'x', lineno=1)])
+                         [Instr('LOAD_DEREF', FreeVar('x'), lineno=1)])
 
     def test_cellvar_freevar(self):
         concrete = ConcreteBytecode()
@@ -257,8 +296,34 @@ class ConcreteBytecodeTests(TestCase):
         bytecode = concrete.to_bytecode()
         self.assertEqual(bytecode.cellvars, ['cell'])
         self.assertEqual(list(bytecode),
-                         [Instr('LOAD_DEREF', 'cell', lineno=1),
-                          Instr('LOAD_DEREF', 'free', lineno=1)])
+                         [Instr('LOAD_DEREF', CellVar('cell'), lineno=1),
+                          Instr('LOAD_DEREF', FreeVar('free'), lineno=1)])
+
+    def test_load_classderef(self):
+        concrete = ConcreteBytecode()
+        concrete.cellvars = ['__class__']
+        concrete.freevars = ['__class__']
+        concrete.extend([ConcreteInstr('LOAD_CLASSDEREF', 1),
+                         ConcreteInstr('STORE_DEREF', 1)])
+
+        bytecode = concrete.to_bytecode()
+        self.assertEqual(bytecode.freevars, ['__class__'])
+        self.assertEqual(bytecode.cellvars, ['__class__'])
+        self.assertEqual(list(bytecode),
+                         [Instr('LOAD_CLASSDEREF', FreeVar('__class__'), lineno=1),
+                          Instr('STORE_DEREF', FreeVar('__class__'), lineno=1)])
+
+        concrete = bytecode.to_concrete_bytecode()
+        self.assertEqual(concrete.freevars, ['__class__'])
+        self.assertEqual(concrete.cellvars, ['__class__'])
+        self.assertEqual(list(concrete),
+                         [ConcreteInstr('LOAD_CLASSDEREF', 1, lineno=1),
+                          ConcreteInstr('STORE_DEREF', 1, lineno=1)])
+
+        code = concrete.to_code()
+        self.assertEqual(code.co_freevars, ('__class__',))
+        self.assertEqual(code.co_cellvars, ('__class__',))
+        self.assertEqual(code.co_code, b'\x94\x01\x00\x89\x01\x00')
 
 
 class ConcreteFromCodeTests(TestCase):
@@ -382,25 +447,27 @@ class BytecodeToConcreteTests(TestCase):
         # x = 7
         # y = 8
         # z = 9
-        code = Bytecode()
-        code.extend([Instr("LOAD_CONST", 7),
-                     Instr("STORE_NAME", 'x'),
-                     SetLineno(2),
-                     Instr("LOAD_CONST", 8),
-                     Instr("STORE_NAME", 'y'),
-                     SetLineno(3),
-                     Instr("LOAD_CONST", 9),
-                     Instr("STORE_NAME", 'z')])
+        concrete = ConcreteBytecode()
+        concrete.consts = [7, 8, 9]
+        concrete.names = ['x', 'y', 'z']
+        concrete.first_lineno = 3
+        concrete.extend([ConcreteInstr("LOAD_CONST", 0),
+                         ConcreteInstr("STORE_NAME", 0),
+                         SetLineno(4),
+                         ConcreteInstr("LOAD_CONST", 1),
+                         ConcreteInstr("STORE_NAME", 1),
+                         SetLineno(5),
+                         ConcreteInstr("LOAD_CONST", 2),
+                         ConcreteInstr("STORE_NAME", 2)])
 
-        concrete = code.to_concrete_bytecode()
-        self.assertEqual(concrete.consts, [7, 8, 9])
-        self.assertEqual(concrete.names, ['x', 'y', 'z'])
-        code.extend([ConcreteInstr("LOAD_CONST", 0, lineno=1),
-                     ConcreteInstr("STORE_NAME", 0, lineno=1),
-                     ConcreteInstr("LOAD_CONST", 1, lineno=2),
-                     ConcreteInstr("STORE_NAME", 1, lineno=2),
-                     ConcreteInstr("LOAD_CONST", 2, lineno=3),
-                     ConcreteInstr("STORE_NAME", 2, lineno=3)])
+        code = concrete.to_bytecode()
+        self.assertEqual(code,
+                         [Instr("LOAD_CONST", 7, lineno=3),
+                          Instr("STORE_NAME", 'x', lineno=3),
+                          Instr("LOAD_CONST", 8, lineno=4),
+                          Instr("STORE_NAME", 'y', lineno=4),
+                          Instr("LOAD_CONST", 9, lineno=5),
+                          Instr("STORE_NAME", 'z', lineno=5)])
 
     def test_extended_jump(self):
         NOP = bytes((opcode.opmap['NOP'],))
@@ -482,8 +549,8 @@ class BytecodeToConcreteTests(TestCase):
         code = Bytecode()
         code.cellvars = ['x']
         code.freevars = ['y']
-        code.extend([Instr('LOAD_DEREF', 'x', lineno=1),
-                     Instr('LOAD_DEREF', 'y', lineno=1)])
+        code.extend([Instr('LOAD_DEREF', CellVar('x'), lineno=1),
+                     Instr('LOAD_DEREF', FreeVar('y'), lineno=1)])
         concrete = code.to_concrete_bytecode()
         self.assertEqual(concrete.cellvars, ['x'])
         self.assertEqual(concrete.freevars, ['y'])
