@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 import opcode
 import unittest
-from bytecode import UNSET, Label, Instr
+from bytecode import UNSET, Label, Instr, CellVar, FreeVar, BasicBlock
 from bytecode.tests import TestCase
 
 
@@ -19,16 +19,54 @@ class InstrTests(TestCase):
         with self.assertRaises(ValueError):
             Instr("xxx")
 
-        # label
+    def test_invalid_arg(self):
         label = Label()
-        with self.assertRaises(ValueError):
-            Instr("LOAD_CONST", label)
+        block = BasicBlock()
+
+        # has_jump()
+        self.assertRaises(TypeError, Instr, "JUMP_ABSOLUTE", 1)
+        self.assertRaises(TypeError, Instr, "JUMP_ABSOLUTE", 1.0)
         Instr("JUMP_ABSOLUTE", label)
+        Instr("JUMP_ABSOLUTE", block)
+
+        # hasfree
+        self.assertRaises(TypeError, Instr, "LOAD_DEREF", "x")
+        Instr("LOAD_DEREF", CellVar("x"))
+        Instr("LOAD_DEREF", FreeVar("x"))
+
+        # haslocal
+        self.assertRaises(TypeError, Instr, "LOAD_FAST", 1)
+        Instr("LOAD_FAST", "x")
+
+        # hasname
+        self.assertRaises(TypeError, Instr, "LOAD_NAME", 1)
+        Instr("LOAD_NAME", "x")
+
+        # hasconst
+        self.assertRaises(ValueError, Instr, "LOAD_CONST")   # UNSET
+        self.assertRaises(ValueError, Instr, "LOAD_CONST", label)
+        self.assertRaises(ValueError, Instr, "LOAD_CONST", block)
+        Instr("LOAD_CONST", 1.0)
+        Instr("LOAD_CONST", object())
+
+        # HAVE_ARGUMENT
+        self.assertRaises(ValueError, Instr, "CALL_FUNCTION", -1)
+        self.assertRaises(TypeError, Instr, "CALL_FUNCTION", 3.0)
+        Instr("CALL_FUNCTION", 3)
+
+        # test maximum argument
+        self.assertRaises(ValueError, Instr, "CALL_FUNCTION", 2147483647 + 1)
+        instr = Instr("CALL_FUNCTION", 2147483647)
+        self.assertEqual(instr.arg, 2147483647)
+
+        # not HAVE_ARGUMENT
+        self.assertRaises(ValueError, Instr, "NOP", 0)
+        Instr("NOP")
 
     def test_attr(self):
         instr = Instr("LOAD_CONST", 3, lineno=5)
         self.assertEqual(instr.name, 'LOAD_CONST')
-        self.assertEqual(instr.op, 100)
+        self.assertEqual(instr.opcode, 100)
         self.assertEqual(instr.arg, 3)
         self.assertEqual(instr.lineno, 5)
 
@@ -36,9 +74,9 @@ class InstrTests(TestCase):
         self.assertRaises(ValueError, setattr, instr, 'lineno', 0)
         self.assertRaises(TypeError, setattr, instr, 'lineno', 1.0)
         self.assertRaises(TypeError, setattr, instr, 'name', 5)
-        self.assertRaises(TypeError, setattr, instr, 'op', 1.0)
-        self.assertRaises(ValueError, setattr, instr, 'op', -1)
-        self.assertRaises(ValueError, setattr, instr, 'op', 255)
+        self.assertRaises(TypeError, setattr, instr, 'opcode', 1.0)
+        self.assertRaises(ValueError, setattr, instr, 'opcode', -1)
+        self.assertRaises(ValueError, setattr, instr, 'opcode', 255)
 
         # arg can take any attribute but cannot be deleted
         instr.arg = -8
@@ -50,11 +88,11 @@ class InstrTests(TestCase):
         self.assertIs(instr.arg, UNSET)
 
     def test_modify_op(self):
-        instr = Instr("LOAD_CONST", 3)
+        instr = Instr("LOAD_NAME", 'x')
         load_fast = opcode.opmap['LOAD_FAST']
-        instr.op = load_fast
+        instr.opcode = load_fast
         self.assertEqual(instr.name, 'LOAD_FAST')
-        self.assertEqual(instr.op, load_fast)
+        self.assertEqual(instr.opcode, load_fast)
 
     def test_extended_arg(self):
         instr = Instr("LOAD_CONST", 0x1234abcd)
@@ -73,29 +111,32 @@ class InstrTests(TestCase):
         self.assertNotEqual(instr, Instr("LOAD_CONST", 3))
         self.assertNotEqual(instr, Instr("LOAD_CONST", 3, lineno=6))
         # different op
-        self.assertNotEqual(instr, Instr("LOAD_FAST", 3, lineno=7))
+        self.assertNotEqual(instr, Instr("LOAD_FAST", 'x', lineno=7))
         # different arg
         self.assertNotEqual(instr, Instr("LOAD_CONST", 4, lineno=7))
 
     def test_has_jump(self):
-        jump = Instr("JUMP_ABSOLUTE", 3)
+        label = Label()
+        jump = Instr("JUMP_ABSOLUTE", label)
         self.assertTrue(jump.has_jump())
 
-        instr = Instr("LOAD_FAST", 2)
+        instr = Instr("LOAD_FAST", 'x')
         self.assertFalse(instr.has_jump())
 
     def test_is_cond_jump(self):
-        jump = Instr("POP_JUMP_IF_TRUE", 3)
+        label = Label()
+        jump = Instr("POP_JUMP_IF_TRUE", label)
         self.assertTrue(jump.is_cond_jump())
 
-        instr = Instr("LOAD_FAST", 2)
+        instr = Instr("LOAD_FAST", 'x')
         self.assertFalse(instr.is_cond_jump())
 
     def test_is_uncond_jump(self):
-        jump = Instr("JUMP_ABSOLUTE", 3)
+        label = Label()
+        jump = Instr("JUMP_ABSOLUTE", label)
         self.assertTrue(jump.is_uncond_jump())
 
-        instr = Instr("POP_JUMP_IF_TRUE", 2)
+        instr = Instr("POP_JUMP_IF_TRUE", label)
         self.assertFalse(instr.is_uncond_jump())
 
     def test_const_key_not_equal(self):
